from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import torch.nn as nn
import logging
import torch
import models.mfvilayers as mfvilayers


def finitialize(modules):
    logging.info("Initializing MSRA")
    for layer in modules:
        if isinstance(layer, (mfvilayers.Conv2dMFMC, mfvilayers.LinearMFMC, nn.Conv2d, nn.Linear)):
            nn.init.kaiming_normal_(layer.weight)
            if layer.bias is not None:
                nn.init.constant_(layer.bias, 0)


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)

def conv3x3Bayes(in_planes, out_planes, stride=1, groups=1, dilation=1, prior_precision=1e0, var_dampening=1e-4):
    """3x3 convolution with padding"""
    return mfvilayers.Conv2dMFMC(in_planes, out_planes, kernel_size=3, stride=stride,
         padding=dilation, groups=groups, bias=False, dilation=dilation,
         prior_precision=prior_precision, var_dampening=var_dampening)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)

def conv1x1Bayes(in_planes, out_planes, stride=1, prior_precision=1e0, var_dampening=1e-4):
    """1x1 convolution"""
    return mfvilayers.Conv2dMFMC(in_planes, out_planes, kernel_size=1, stride=stride, bias=False,
        prior_precision=prior_precision, var_dampening=var_dampening)


class BasicBlock(nn.Module):
    expansion = 1
    __constants__ = ['downsample']

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class BasicBlockEncoder(nn.Module):
    expansion = 1
    __constants__ = ['downsample']

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None, l2_norm=False):
        super(BasicBlockEncoder, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride
        self.planes = planes
        self.l2_norm = l2_norm

    def forward(self, x):
        if self.l2_norm:
            x = x / torch.sqrt(x.pow(2).sum(dim=(1,2,3)) + 1e-5).reshape(-1,1,1,1)
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        z = out[:, :self.planes // 2, :, :]
        z_var = out[:, self.planes // 2:, :, :]

        return z, z_var


class BasicBlockDecoder(nn.Module):
    expansion = 1
    __constants__ = ['downsample']

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None, prior_precision=1e0, var_dampening=1e-4):
        super(BasicBlockDecoder, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        self.conv1 = conv3x3Bayes(inplanes, planes, stride, prior_precision=prior_precision, var_dampening=var_dampening)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3Bayes(planes, planes, prior_precision=prior_precision, var_dampening=var_dampening)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class ResNetNCVIBNBase(nn.Module):

    def __init__(self, block, layers, num_classes=100, prior_precision=1e0, var_dampening=1e-4, zero_init_residual=False,
                 groups=1, width_per_group=64,  replace_stride_with_dilation=None, l2_norm=False, maxpool=False):
        super(ResNetNCVIBNBase, self).__init__()
        self._norm_layer = nn.BatchNorm2d
        self._maxpool = maxpool

        self.inplanes = 64
        self.dilation = 1
        if replace_stride_with_dilation is None:
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError("replace_stride_with_dilation should be None "
                             "or a 3-element tuple, got {}".format(replace_stride_with_dilation))
        self.groups = groups
        self.base_width = width_per_group
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=3, stride=1, padding=1,
                               bias=False)
        self.bn1 = self._norm_layer(self.inplanes)
        self.relu = nn.ReLU()
        if self._maxpool:
            self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2,
                                       dilate=replace_stride_with_dilation[0])
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2,
                                       dilate=replace_stride_with_dilation[1])
        self.layer4 = self._make_layer_encoder(block, 512, layers[3], stride=2,
                                       dilate=replace_stride_with_dilation[2], l2_norm=l2_norm)
        self.decoder_block = self._make_layer_decoder(block, 512, stride=2,
                                       dilate=replace_stride_with_dilation[2],
                                       prior_precision=prior_precision, var_dampening=var_dampening)
        self.avgpool = nn.AdaptiveAvgPool2d((1,1))
        if block == 'BasicBlock':
            block = BasicBlock
        self.decoder_fc = mfvilayers.LinearMFMC(256 * block.expansion, num_classes, 
            var_dampening=var_dampening, prior_precision=prior_precision)

        finitialize(self.modules())

    def _make_layer(self, block_name, planes, blocks, stride=1, dilate=False):
        if block_name == 'BasicBlock':
            block = BasicBlock
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer))

        return nn.Sequential(*layers)

    def _make_layer_encoder(self, block_name, planes, blocks, stride=1, dilate=False, l2_norm=False):
        if block_name == 'BasicBlock':
            block = BasicBlock
            block_enc = BasicBlockEncoder
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        if blocks > 2:
            layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                                self.base_width, previous_dilation, norm_layer))
            self.inplanes = planes * block.expansion
            for _ in range(1, blocks - 2):
                layers.append(block(self.inplanes, planes, groups=self.groups,
                                    base_width=self.base_width, dilation=self.dilation,
                                    norm_layer=norm_layer))
            layers.append(block_enc(self.inplanes, planes, groups=self.groups,
                                    base_width=self.base_width, dilation=self.dilation,
                                    norm_layer=norm_layer, l2_norm=l2_norm))
        else:
            layers.append(block_enc(self.inplanes, planes, stride, downsample, groups=self.groups,
                                    base_width=self.base_width, dilation=self.dilation,
                                    norm_layer=norm_layer, l2_norm=l2_norm))

        return nn.Sequential(*layers)

    def _make_layer_decoder(self, block_name, planes, stride=1, dilate=False, prior_precision=1e0, var_dampening=1e-4):
        if block_name == 'BasicBlock':
            block = BasicBlock
            block_dec = BasicBlockDecoder
        norm_layer = self._norm_layer
        if dilate:
            self.dilation *= stride
            stride = 1
            
        layers = []
        self.inplanes = planes // 2 * block.expansion
        layers.append(block_dec(self.inplanes, planes // 2, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer, prior_precision=prior_precision, var_dampening=var_dampening))

        return nn.Sequential(*layers)

    def _forward_impl(self, inputs_mean):
        x = self.conv1(inputs_mean)
        x = self.bn1(x)
        x = self.relu(x)
        if self._maxpool:
            x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        z, z_var = self.layer4(x)
        
        return z, z_var

    def forward(self, inputs_mean):
        return self._forward_impl(inputs_mean)

    def decoder(self, z):
        z = self.decoder_block(z)
        z = self.avgpool(z)
        z = torch.flatten(z, 1)
        z = self.decoder_fc(z)
        return z

    def kl_div(self):
        kl = 0.0
        for module in self.modules():
            if isinstance(module, (mfvilayers.LinearMFMC, mfvilayers.Conv2dMFMC)):
                kl += module.kl_div()
        return kl